#! /usr/bin/env python

"""
For the command line interface. Not finished.

Pompei, photo mosaic creations with python.

Usage:
  python pompei [options]
  
Options:
  --moviefile=<filename>
  --folder=<foldername>    
  --movie_fps=<fps>      
  --movieframe=<filename> <time>
  --signature=<sx> <sy>
  --resulution=<rx> <ry>
  --help -h              Show this screen.
  
"""